var utils = require("./_utils.js");
var iterateObject = require("./_iterateObject.js");
var parser = require("./_parser.js");
var apply = require("./_apply.js");
var jsonDotNetDate = require("json-dotnet-date")();
var iso8601Date = require("iso8601-date");
var merge = require("just-extend")

var defaultOptions = {
    arrayOpenChar: "(",
    arrayCloseChar: ")",
    propAccessChar: ".",
    setValueChar: "=",
    queryDelimiterChar: "&",
    useDotNetDateFormat: false,
    lowercaseNames: false,
    formatters: [
        {
            accept: iso8601Date.testObject,
            format: iso8601Date.format
        }
    ],
    parsers: [
        {
            accept: function (str) {
                var matchInteger = /^\d+$/i;
                var matchFloat = /^\d+(?:\.\d+)$/i;
                return matchInteger.test(str) || matchFloat.test(str);
            },
            parse: function (str) {
                var matchInteger = /^\d+$/i;
                var matchFloat = /^\d+(?:\.\d+)$/i;

                if (matchFloat.exec(str)) {
                    return Number.parseFloat(str);
                }
                if (matchInteger.exec(str)) {
                    return Number.parseInt(str);
                }


                throw "Unexpected value: " + str;
            }
        },
        {
            accept: iso8601Date.testStr,
            parse: iso8601Date.parse
        },
        {
            accept: jsonDotNetDate.testStr,
            parse: jsonDotNetDate.parse
        }
    ]
};

function constructor(opts) {
    var opts = merge(true, {}, defaultOptions, opts || {});

    if (opts.useDotNetDateFormat) {
        opts.formatters[0] = {
            accept: jsonDotNetDate.testObject,
            format: jsonDotNetDate.format,
        };
    }

    function decode(query, to) {
        var tokens = parser.splitQuery(query, opts.queryDelimiterChar, opts.setValueChar);
        for (var i = 0; i < tokens.length; i++) {
            var token = tokens[i];
            var path = parser.splitPath(token[0]);
            var value = token[1];

            if (opts.lowercaseNames) {
                for(var pIndex=0;pIndex<path.length;pIndex++){
                    path[pIndex] = Number.isInteger(path[pIndex])?path[pIndex]:path[pIndex].toLowerCase();
                }
            }

            //assume target type by first path
            if (!to) {
                to = Number.isInteger(path[0]) ? [] : {};
            }

            value = value.replace("+", "%20");
            value = decodeURIComponent(value);

            for (var parserIndex = 0; parserIndex < opts.parsers.length; parserIndex++) {
                var p = opts.parsers[parserIndex];
                if (p.accept(value)) {
                    value = p.parse(value);
                    break;
                }
            }

            apply(to, path, value)
        }

        return to;
    };


    function _toNameValuePairs(obj) {
        var pairs = [];
        iterateObject(obj, function (node, start) {
            if (!start) {
                return;
            }

            var isLeaf = false;
            var value = node.value;
            // detect leaf
            isLeaf = node.type == "simple";
            if (!isLeaf) {
                for (var fmtIndex = 0; fmtIndex < opts.formatters.length; fmtIndex++) {
                    var f = opts.formatters[fmtIndex];
                    if (f.accept(value)) {
                        value = f.format(value);
                        isLeaf = true;
                        break;
                    }
                }
            }

            if (!isLeaf) {
                return;
            }

            value = encodeURIComponent(value);
            value = value.replace(/%20/g, "+");

            var nodesPath = [];
            var current = node;
            while (current != null) {
                if (current == null) {
                    break;
                }
                nodesPath.push(current);
                current = current.parent
            }

            var name = "";
            for (var i = nodesPath.length - 1; i >= 0; i--) {
                var node = nodesPath[i];
                if (!node.parent) {
                    continue;
                }

                var parent = node.parent;

                switch (parent.type) {
                    case "array":
                        name += opts.arrayOpenChar + node.id + opts.arrayCloseChar;
                        break;
                    case "object":
                        name += name.length > 0 ? opts.propAccessChar : "";
                        name += node.id;
                        break;

                    default:
                        throw "Unexpected node type: " + parent.type;
                }
            }

            if (opts.lowercaseNames) {
                name = name.toLowerCase();
            }

            pairs.push(name + opts.setValueChar + value)
        })

        return pairs;
    }

    function encode(object) {
        var pairs = _toNameValuePairs(object);
        return pairs.join(opts.queryDelimiterChar);
    }

    return {
        decode: decode,
        encode: encode,
    };
};

constructor.defaultOptions = defaultOptions;
module.exports = constructor;